/*
Copyright (C) 2003 Bj\"orn Hoffmeister

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package de.uni_luebeck.tcs.matlab;

/**
* This is a translation of a FORTRAN program by W. J. Cody,
* Argonne National Laboratory, NETLIB/SPECFUN, March 19, 1990.
* The main computation evaluates near-minimax approximations
* from "Rational Chebyshev approximations for the error function"
* by W. J. Cody, Math. Comp., 1969, PP. 631-638.
*/

public class ErrorFunction {
	// Konstantendeklaration
	private static final double[] a = {
		3.16112374387056560e00, 1.13864154151050156e02,
		3.77485237685302021e02, 3.20937758913846947e03,
		1.85777706184603153e-1
	};
	private static final double[] b = {
		2.36012909523441209e01, 2.44024637934444173e02,
		1.28261652607737228e03, 2.84423683343917062e03
	};
	private static final double[] c = {
		5.64188496988670089e-1, 8.88314979438837594e00,
		6.61191906371416295e01, 2.98635138197400131e02,
		8.81952221241769090e02, 1.71204761263407058e03,
		2.05107837782607147e03, 1.23033935479799725e03,
		2.15311535474403846e-8
	};
	private static final double[] d = {
		1.57449261107098347e01, 1.17693950891312499e02,
		5.37181101862009858e02, 1.62138957456669019e03,
		3.29079923573345963e03, 4.36261909014324716e03,
		3.43936767414372164e03, 1.23033935480374942e03
	};
	private static final double[] p = {
		3.05326634961232344e-1, 3.60344899949804439e-1,
		1.25781726111229246e-1, 1.60837851487422766e-2,
		6.58749161529837803e-4, 1.63153871373020978e-2
	};
	private static final double[] q = {
		2.56852019228982242e00, 1.87295284992346047e00,
		5.27905102951428412e-1, 6.05183413124413191e-2,
		2.33520497626869185e-3
	};
	private static final double[] inva = {
		0.886226899, -1.645349621,
		0.914624893, -0.140543331
	};
	private static final double[] invb = {
		-2.118377725, 1.442710462,
		-0.329097515, 0.012229801
	};
	private static final double[] invc = {
		-1.970840454, -1.624906493,
		3.429567803, 1.641345311
	};
	private static final double[] invd = {
		3.543889200, 1.637067800
	};	
	private static final double xbreak1 = 0.46875;
	private static final double xbreak2 = 4.0;
	private static final double y0 = 0.7;

	/**
	* ERFCORE Core algorithm for error functions.
	*   erf(x) = erfcore(x,0)
	*   erfc(x) = erfcore(x,1)
	*   erfcx(x) = exp(x^2)*erfc(x) = erfcore(x,2)
	*/
	private static double erfcore(double x, int jint) {
		double result = java.lang.Double.NaN;
		double y = Math.abs(x);
		if (y <= xbreak1) {
			double z = y * y;
			double xnum = a[4] * z;
			double xden = z;
			for (int i=0; i < 3; i++) {
				xnum = (xnum + a[i]) * z;
				xden = (xden + b[i]) * z;
			}
			result = x * (xnum + a[3]) / (xden + b[3]);
			if (jint != 0) result = 1 - result;
			if (jint == 2) result = Math.exp(z) * result;
		} else if ((y > xbreak1) && (y <= xbreak2)) {
			double xnum = c[8] * y;
			double xden = y;
			for (int i=0; i < 7; i++) {
				xnum = (xnum + c[i]) * y;
				xden = (xden + d[i]) * y;
			}
			result = (xnum + c[7]) / (xden + d[7]);
			if (jint != 2) {
				double z = Round.fix(y * 16.0) / 16.0;
				double del = (y - z) * (y + z);
				result = Math.exp(-z * z) * Math.exp(-del) * result;
			}
		} else {
			double z = 1.0 / (y * y);
			double xnum = p[5] * z;
			double xden = z;
			for (int i=0; i < 4; i++) {
				xnum = (xnum + p[i]) * z;
				xden = (xden + q[i]) * z;
			}
			result = z * (xnum + p[4]) / (xden + q[4]);
			if (jint != 2) {
				z = Round.fix(y * 16.0) / 16.0;
				double del = (y - z) * (y + z);
				result = Math.exp(-z * z) * Math.exp(-del) * result;
				if (java.lang.Double.isNaN(result) || java.lang.Double.isInfinite(result)) result = 0;
			}
		}		
		if (jint == 0) {
			if (x > xbreak1) result = (0.5 - result) + 0.5;
			if (x < -xbreak1) result = (-0.5 + result) - 0.5;
		} else if (jint == 1) {
			if (x < -xbreak1) result = 2.0 - result;
		} else {
			if (x < -xbreak1) {
				double z = Round.fix(x * 16.0) / 16.0;
				double del = (x - z) * (x + z);
				y = Math.exp(z * z) * Math.exp(del);
				result = (y + y) - result;
			}
		}	
		return result;
	}

	/**
	* ERF Error function.
    	*   Y = ERF(X) is the error function for each element of X.  X must be
    	*   real. The error function is defined as:
      	*  erf(x) = 2/sqrt(pi) * integral from 0 to x of exp(-t^2) dt.
	*/
	public static double erf(double x) { return erfcore(x, 0); }

 	/**
 	* ERFC  Complementary error function.
    	*   Y = ERFC(X) is the complementary error function for each element
    	*   of X.  X must be real.  The complementary error function is
    	*   defined as:
      	*   erfc(x) = 2/sqrt(pi) * integral from x to inf of exp(-t^2) dt.
        *      = 1 - erf(x).	
        */
	public static double erfc(double x) { return erfcore(x, 1); }

	/**
	* ERFCX  Scaled complementary error function.
	*   Y = ERFCX(X) is the scaled complementary error function for each
	*   element of X.  X must be real.  The scaled complementary error
	*   function is defined as: 
	*     erfcx(x) = exp(x^2) * erfc(x)
	*/
	public static double erfcx(double x) { return erfcore(x, 2); }

	/**
	* ERFINV Inverse error function.
	*   X = ERFINV(Y) is the inverse error function for each element of X.
	*   The inverse error functions satisfies y = erf(x), for -1 <= y < 1
	*   and -inf <= x <= inf.
	*/
	public static double erfinv(double y) {
		double x = java.lang.Double.NaN;
		// Central range
		if (Math.abs(y) <= y0) {
			double z = y * y;
			x = y * (((inva[3] * z + inva[2]) * z + inva[1]) * z + inva[0]) /
				((((invb[3] * z + invb[2]) * z + invb[1]) * z + invb[0]) * z + 1);
		// Near end points range
		} else if ((y0 < y) && (y < 1.0)) {
			double z = Math.sqrt(-Math.log((1 - y) / 2));
			x = (((invc[3] * z + invc[2]) * z + invc[1]) * z + invc[0]) / ((invd[1] * z + invd[0]) * z + 1);
		} else if ((-y0 > y) && (y > -1)) {
			double z = Math.sqrt(-Math.log((1 + y) / 2));
			x = -(((invc[3] * z + invc[2]) * z + invc[1]) * z + invc[0]) / ((invd[1] * z + invd[0]) * z + 1);
		}
		// Exceptional cases
		if (y == -1) {
			x = java.lang.Double.NEGATIVE_INFINITY;
		} else if (y == 1) {
			x = java.lang.Double.POSITIVE_INFINITY;
		// Two steps of Newton-Raphson correction to full accuracy.
		// Without these steps, erfinv(y) would be about 3 times
		// faster to compute, but accurate to only about 6 digits.
		} else if (Math.abs(y) <= 1) {
			x = x - (erf(x) - y) / (2 / Math.sqrt(Math.PI) * Math.exp(-x * x));
			x = x - (erf(x) - y) / (2 / Math.sqrt(Math.PI) * Math.exp(-x * x));
		}
		return x;
	}
}
	

