/*
Copyright (C) 2003 Bj\"orn Hoffmeister

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package de.uni_luebeck.tcs.matlab;

public class NormalDist {
	
	/**
	*NORMPDF Normal probability density function (pdf).
	*   Y = NORMPDF(X,MU,SIGMA) Returns the normal pdf with mean, MU, 
	*   and standard deviation, SIGMA, at the values in X. 
	*
	*   The size of Y is the common size of the input arguments. A scalar input  
	*   functions as a constant matrix of the same size as the other inputs.     
	*
	*   Default values for MU and SIGMA are 0 and 1 respectively.
	*
	*   References:
	*      [1]  M. Abramowitz and I. A. Stegun, "Handbook of Mathematical
	*      Functions", Government Printing Office, 1964, 26.1.26.
	*/
	public static double normPDF(double x) { return normPDF(x, 0.0, 1.0); }
	
	public static double normPDF(double x, double mu) { return normPDF(x, mu, 1.0); }	
	
	public static double normPDF(double x, double mu, double sigma) {
		double y = 0.0;
		// Return NaN if SIGMA is not positive.
		if (sigma <= 0) return java.lang.Double.NaN;
		// Calculate PDF 
		double xn = (x - mu) / sigma;
		y = Math.exp(-0.5 * (xn * xn)) / (Math.sqrt(2.0 * Math.PI) * sigma);
		return y;
	}

	/**
	*NORMCDF Normal cumulative distribution function (cdf).
	*   P = NORMCDF(X,MU,SIGMA) computes the normal cdf with mean MU and
	*   standard deviation SIGMA at the values in X.
	*
	*   The size of P is the common size of X, MU and SIGMA. A scalar input  
	*   functions as a constant matrix of the same size as the other inputs.    
	*
	*   Default values for MU and SIGMA are 0 and 1 respectively.
	*
	*   References:
	*      [1]  M. Abramowitz and I. A. Stegun, "Handbook of Mathematical
	*      Functions", Government Printing Office, 1964, 26.2.
	*/
	public static double normCDF(double x) { return normCDF(x, 0.0, 1.0); }
	
	public static double normCDF(double x, double mu) { return normCDF(x, mu, 1.0); }	
	
	public static double normCDF(double x, double mu, double sigma) {
		double p = 0.0;
		// Return NaN if SIGMA is not positive.
		if (sigma <= 0) return java.lang.Double.NaN;
		// Express normal CDF in terms of the error function.
		if (sigma > 0) p = 0.5 * ErrorFunction.erfc(-(x - mu) / (sigma * Math.sqrt(2.0)));
		// Make sure that round-off errors never make P greater than 1.
		return ((p > 1)?1.0:p);
	}
	
	/**
	*NORMINV Inverse of the normal cumulative distribution function (cdf).
	*   X = NORMINV(P,MU,SIGMA) finds the inverse of the normal cdf with
	*   mean, MU, and standard deviation, SIGMA.
	*
	*   The size of X is the common size of the input arguments. A scalar input  
	*   functions as a constant matrix of the same size as the other inputs.    
	*
	*   Default values for MU and SIGMA are 0 and 1 respectively.
	*
	*   References:
	*      [1]  M. Abramowitz and I. A. Stegun, "Handbook of Mathematical
	*      Functions", Government Printing Office, 1964, 7.1.1 and 26.2.2
	*/
	public static double norminv(double p) { return norminv(p, 0.0, 1.0); }

	public static double norminv(double p, double mu) { return norminv(p, mu, 1.0); }
	
	public static double norminv(double p, double mu, double sigma) {
		double x = 0.0;
		// Return NaN if the arguments are outside their respective limits.
		if ((sigma <= 0.0) || (p < 0.0) || (p > 1.0)) return java.lang.Double.NaN;
		// Put in the correct values when P is either 0 or 1.
		if (p == 0.0) {
			x = java.lang.Double.NEGATIVE_INFINITY;
		} else if (p == 1.0) {
			x = java.lang.Double.POSITIVE_INFINITY;
		} else {
			x = Math.sqrt(2) * sigma * ErrorFunction.erfinv(2 * p - 1) + mu;
		}
		return x;
	}
	
}

