/*
Copyright (C) 2003 Bj\"orn Hoffmeister

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package de.uni_luebeck.tcs.toolbox.duembgen;

import de.uni_luebeck.tcs.matlab.*;
import de.uni_luebeck.tcs.visualization.Debug;

/**
 * [index,z] = TautString(x,y_low,y_upp);
 * computes the knots of the "taut string" connecting 
 * (x(1),a) and (x(n),b), where x is a vector with 
 * components x(1) < x(2) < ... < x(n), y_low and y_upp 
 * are two other vectors of length n, where
 *    y_low(1) = y_upp(1) = a ,
 *    y_low(n) = y_upp(n) = b ,
 *    y_low(i) <= y_upp(i) for i=2,3,...,n-1 .
 * The knots of the taut string are given by
 *    (x(index(i)), z(i)), 1 <= i <= length(index).
 * 
 * Sources:
 * - E. Mammen and S. van de Geer (1995). 
 *   Locally adaptive regression splines. 
 *   Annals of Statistics 25, 387-413.
 * - P.L. Davies and A. Kovac (2001).
 *   Local Extremes, runs, strings and multiresolution. 
 *   to appear in the Annals of Statistics
 * 
 * Lutz Duembgen, November 24, 2000
**/

public class TautString {
//constants -----------------------------------------------------------------------------------------------------------
	//private
	//protected
	//public

//inner classes -------------------------------------------------------------------------------------------------------
	
//attributes ----------------------------------------------------------------------------------------------------------
	//private
	private int[] intIndex = null;
	private double[] dblX = null;
	private double[] dblZ = null;

	private Debug d = new Debug("TautString");
	
	//protected
	//public

//constructors --------------------------------------------------------------------------------------------------------
	//private
	//protected
	//public
	public TautString(double[] dblX, double[] dblYLow, double[] dblYUp) throws IllegalArgumentException {
		if ((dblX == null) || (dblYLow == null) || (dblYUp == null))
			throw new IllegalArgumentException("dblX, dblYLow or dblYUp must not be null");			
		if (dblX.length == 0)
			throw new IllegalArgumentException("dblX, dblYLow or dblYUp must not be empty");		
		if ((dblX.length != dblYLow.length) || (dblX.length != dblYUp.length))
			throw new IllegalArgumentException("dblX, dblYLow and dblYUp must have same length");
		calcTautString(dblX, dblYLow, dblYUp);
	}
	

//methods -------------------------------------------------------------------------------------------------------------
	protected void calcTautString(double[] dblX, double[] dblYLow, double[] dblYUp) {
		int intN = dblX.length;		

	// init lower string, upper string, taut string
		int[] intIndexLow = new int[intN];
		double[] dblSlopeLow = new double[intN];
		int[] intIndexUp = new int[intN];
		double[] dblSlopeUp = new double[intN];		
		int[] intIndex = new int[intN];
		double[] dblZ = new double[intN];				
		for (int i = 0; i < intN; i++) {
			intIndexLow[i] = 0; dblSlopeLow[i] = 0.0;
			intIndexUp[i] = 0; dblSlopeUp[i] = 0.0;
			intIndex[i] = 0; dblZ[i] = 0.0;
		}
		dblSlopeLow[0] = Double.POSITIVE_INFINITY;
		int intSLow = 0;
		int intCLow = 0;
		dblSlopeUp[0] = Double.NEGATIVE_INFINITY;
		int intSUp = 0;
		int intCUp = 0;
		dblZ[0] = dblYLow[0];
		int intC = 0;		
		
	// build taut string
		for (int j = 1; j < intN; j++) {
		// append new knot, {j}, to list index_low:
			intCLow++;
			intIndexLow[intCLow] = j;
			dblSlopeLow[intCLow] = (dblYLow[j] - dblYLow[j - 1]) / (dblX[j] - dblX[j - 1]);
		// check slope_low for antitonicity:
			while (
				(intCLow > intSLow + 1) && 
				(dblSlopeLow[Math.max(intSLow, intCLow - 1)] <= dblSlopeLow[intCLow])
			) {
				// ("pool adjacent violators") 
				// knot number c_low-1 is removed:
				intCLow--;
				intIndexLow[intCLow] = j;
				if (intCLow > intSLow + 1) {
					dblSlopeLow[intCLow] = 
						(dblYLow[j] - dblYLow[intIndexLow[intCLow - 1]]) /
						(dblX[j] - dblX[intIndexLow[intCLow - 1]]);
				} else {
					dblSlopeLow[intCLow] =
						(dblYLow[j] - dblZ[intC]) /
						(dblX[j] - dblX[intIndex[intC]]);
				}
			}

		// append new knot, {j}, to list index_upp:
			intCUp++;
			intIndexUp[intCUp] = j;
			dblSlopeUp[intCUp] = (dblYUp[j] - dblYUp[j - 1]) / (dblX[j] - dblX[j - 1]);
		// check slope_up for isotonicity:
			while (
				(intCUp > intSUp + 1) && 
				(dblSlopeUp[Math.max(intCUp - 1, intSUp)] >= dblSlopeUp[intCUp])
			) {
				// ("pool adjacent violators") 
				// knot number c_up-1 is removed:
				intCUp--;
				intIndexUp[intCUp] = j;
				if (intCUp > intSUp + 1) {
					dblSlopeUp[intCUp] = 
						(dblYUp[j] - dblYUp[intIndexUp[intCUp - 1]]) /
						(dblX[j] - dblX[intIndexUp[intCUp - 1]]);
				} else {
					dblSlopeUp[intCUp] =
						(dblYUp[j] - dblZ[intC]) /
						(dblX[j] - dblX[intIndex[intC]]);
				}
			}		

		// check whether slope_low(s_low+1) <= slope_upp(s_upp+1):
			while (
				(intCLow == intSLow + 1) && 
				(intCUp > intSUp + 1) && 
				(dblSlopeLow[intCLow] >= dblSlopeUp[intSUp + 1])
			) {
				// Remove first knot of upper string:
				intSUp++;
				// Add a knot to the taut string:
				intC++;
				intIndex[intC] = intIndexUp[intSUp];
				dblZ[intC] = dblYUp[intIndex[intC]];
				// Shift first knot of the lower string
				intIndexLow[intSLow] = intIndex[intC];
				dblSlopeLow[intCLow] = 
					(dblYLow[j] - dblZ[intC]) /
					(dblX[j] - dblX[intIndex[intC]]);
			}

			while (
				(intCUp == intSUp + 1) && 
				(intCLow > intSLow + 1) && 
				(dblSlopeUp[intCUp] <= dblSlopeLow[intSLow + 1])
			) {
				// Remove first knot of lower string:
				intSLow++;
				// Add a knot to the taut string:
				intC++;
				intIndex[intC] = intIndexLow[intSLow];
				dblZ[intC] = dblYLow[intIndex[intC]];
				// Shift first knot of the upper string
				intIndexUp[intSUp] = intIndex[intC];
				dblSlopeUp[intCUp] = 
					(dblYUp[j] - dblZ[intC]) /
					(dblX[j] - dblX[intIndex[intC]]);
			}

		}
		
	// Augment taut string by lower (or upper) string:
		for (int i = 0; i < (intCLow - intSLow); i++) {
			intIndex[intC + 1 + i] = intIndexLow[intSLow + 1 + i];
			dblZ[intC + 1 + i] = dblYLow[intIndexLow[intSLow + 1 + i]];
		}
	// Remove superfluous components:
		intC += (intCLow - intSLow);
		this.intIndex = new int[intC + 1];
		System.arraycopy(intIndex, 0, this.intIndex, 0, intC + 1);
		this.dblZ = new double[intC + 1];
		System.arraycopy(dblZ, 0, this.dblZ, 0, intC + 1);
		this.dblX = new double[intC + 1];
		for (int i = 0; i < intC + 1; i++) this.dblX[i] = dblX[this.intIndex[i]];
	}		
	
	//public

// interface ----------------------------------------------------------------------------------------------------------

//listener (public) ---------------------------------------------------------------------------------------------------

//getter-/setter-methods (public) -------------------------------------------------------------------------------------
	public int[] getIndex() { return this.intIndex; }
	
	public double[] getX() { return this.dblX; }
	
	public double[] getZ() { return this.dblZ; }
}
